# -*- coding: utf-8 -*-
import base64
import datetime
import hashlib
import os
import requests
import json
import shutil
import subprocess
import sys
import tempfile
import time
import logging
import random

try:
    import urllib3
except ImportError:
    from requests.packages import urllib3
if hasattr(urllib3, 'disable_warnings'):  # noqa
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)  # noqa

try:
    from const import BRANCH_ALIASES, TIMEOUT, TMPMIGDIR, UPGRADE, UPGRADE_KEYSALT, UPGRADE_LOCATION, UPGRADE_SAAS_KEY, WEBSITE
    from auth_util import SaasAuth
except ImportError:
    from .const import BRANCH_ALIASES, TIMEOUT, TMPMIGDIR, UPGRADE, UPGRADE_KEYSALT, UPGRADE_LOCATION, UPGRADE_SAAS_KEY, WEBSITE
    from .auth_util import SaasAuth

KEY_PATH = os.path.expanduser('~/.ssh/id_rsa')
PY3 = sys.version_info[0] == 3
SUBPROC_TIMEOUT = 3600 * 6

_logger = logging.getLogger(__name__)


def retry(retries, msg, catch=(Exception,), log_errors=True):
    assert retries >= 0

    def decorator(func):

        def wrapper(*args, **kwargs):
            for attempt in range(0, retries + 1):
                try:
                    return func(*args, **kwargs)
                except catch:
                    if log_errors:
                        _logger.exception(msg)
                    if attempt != retries:
                        time.sleep(random.uniform(0.0, 2 ** attempt))
            raise UpgradeError(msg)

        return wrapper

    return decorator


class UpgradeError(Exception):
    pass


class UpgradeAPI():
    _dump_format = 'directory'

    def __init__(self, db, pg_con, pg_env):
        self.endpoint = UPGRADE
        if isinstance(pg_con, str):
            self.pg_con = pg_con.split()
        elif isinstance(pg_con, (tuple, list)):
            self.pg_con = pg_con
        self.pg_env = pg_env
        if db:
            self.db_uuid = db.uuid
            self.source_uuid = db.source_uuid.uuid if db.source_uuid else False

    def new_request(self, db_name, target, actuator, mode='copy', email=None):
        aim = 'production' if mode == 'self' else 'test'
        quiet = mode in ('rr-preflight', 'rr-preflight-muted')
        request_params = self._create_request(db_name, actuator, target, aim, email=email, quiet=quiet)
        self._upload_database(db_name, request_params)
        self._process_request(request_params)
        return {"key": request_params["token"], "request": request_params["request_id"]}

    @retry(5, "Cannot get status")
    def request_status(self, request_id, token):
        """ get upgrade status and reduce it in one of 3 possible statuses:
            running, testing or done
        """
        url = self.endpoint + '/upgrade/request/status'
        res = requests.get(url, json={"params": {"token": token}}, auth=SaasAuth(), timeout=TIMEOUT)
        res.raise_for_status()
        if "result" not in res.json():
            raise UpgradeError(
                "Cannot get status for token %s, invalid response from upgrade server: %s" % (token, res.text)
            )
        response = res.json()["result"]
        if response.get('archived'):
            return {'status': 'archived', 'write_date': response.get('write_date', datetime.datetime.now().isoformat())}
        req_status = response["status"]
        req_reason = response.get('reason', '')

        queued_states = ('new', 'pending')
        running_states = ('progress',)
        failed_states = ('failed',)
        cancelled_states = ('cancelled',)
        success_states = ('done',)
        warning_states = ('warning',)

        status = False
        if req_status in success_states:
            status = 'done'
        if req_status in warning_states:
            status = 'warning'
        if req_status in queued_states + running_states:
            status = 'running'
        if req_status in failed_states:
            status = 'failed'
        if req_status in cancelled_states:
            status = 'cancelled'
        return {
            'status': status or 'failed',
            'elapsed_time': response.get('elapsed_time', 0),
            'disabled_view_count': response.get('disabled_view_count', 0),
            'host_uri': response.get('host_uri', self.endpoint.split('//')[1]),
            'write_date': response.get('write_date', datetime.datetime.now().isoformat()),
            'reason': req_reason,
        }

    def download_database(self, request_id, token, archfile):
        res = self.request_status(request_id, token)
        self._transfer(token, "download", archfile, res['host_uri'])

    def _create_request(self, db_name, actuator, target, aim='test', email=None, quiet=False):
        if not email:
            email = 'migration-saas@odoo.com'

        with open(KEY_PATH + ".pub") as key:
            ssh_key = key.read()

        sub_code = self._get_contract()
        create_req_url = self.endpoint + '/upgrade/request/create'
        req_params = {
            'aim': aim,
            'actuator': actuator,
            'email': email,
            'contract': sub_code,
            'db_uuid': self.db_uuid,
            'target': target.replace('-', '~'),
            'ssh_key': ssh_key,
            'quiet': quiet,
            'api_version': "0.2",
        }
        if actuator == 'meta':
            req_params['env_vars'] = ['ODOO_MIG_TRY_FIX_VIEWS=1']
        create_req_res = requests.post(create_req_url, json={"params": req_params}, auth=SaasAuth(), timeout=TIMEOUT)
        try:
            res = create_req_res.json()
        except ValueError:
            # No json readable, get error
            error = "Unknown Error"
            try:
                create_req_res.raise_for_status()
            except requests.HTTPError as e:
                error = repr(e)
            raise UpgradeError(
                "Invalid response from upgrade server for database %s: %s" % (db_name, error)
            )

        error = res.get("error") or res.get("result", {}).get("error")
        if error:
            msg = repr(error)
            raise UpgradeError("Cannot request migration for database %s: %s" % (db_name, msg))

        result = res["result"]
        request_params = {
            'request_id': result['request_id'],
            'token': result['token'],
            'host_uri': result.get('assigned_host_uri', self.endpoint.split('//')[1]),
        }
        return request_params

    def _upload_database(self, db_name, request_params):
        dump_dir = tempfile.mkdtemp(dir=TMPMIGDIR)
        try:
            dump_file = os.path.join(dump_dir, 'origin.dump')
            pg_dump_cmd = ["pg_dump", "-O", "-Fd", "-j4", "-f", dump_file, db_name]
            p = subprocess.Popen(pg_dump_cmd, env=self.pg_env)
            if not PY3:
                p.wait()
            else:
                p.wait(timeout=SUBPROC_TIMEOUT)
            self._transfer(request_params["token"], "upload", dump_file, request_params["host_uri"])
        finally:
            shutil.rmtree(dump_dir)

    def _process_request(self, request_params):
        process_req_url = 'https://%s/upgrade/request/process' % request_params['host_uri']
        requests.post(process_req_url, json={"params": {"token": request_params["token"]}})

    @retry(5, "Cannot transfer database")
    def _transfer(self, token, transfer_type, dump_file, host_uri=None):
        if not host_uri:
            host_uri = self.endpoint.split('//')[1]

        res = requests.post("https://%s/upgrade/request/transfer/start" % host_uri, json={"params": {
            "token": token,
            "transfer_type": transfer_type
        }}, timeout=TIMEOUT)
        res.raise_for_status()
        if "result" not in res.json():
            raise UpgradeError(
                "Cannot transfer database for token %s, invalid response from upgrade server: %s" % (token, res.text)
            )
        result = res.json()["result"]
        port = result["ssh_port"]
        ssh = "ssh -p %s -o IdentitiesOnly=yes -i %s" % (port, KEY_PATH)
        remote = "odoo@%s:/data" % host_uri
        cmd = ["rsync", "--chmod=u+rwx,g+rwx,o+r", "-arze", ssh]
        if transfer_type == "upload":
            cmd.extend([dump_file, remote])
        else:
            cmd.extend([remote + "/" + result["dump_name"], remote + "/" + "filestore", dump_file])
        if not PY3:
            subprocess.check_call(cmd)
        else:
            subprocess.check_call(cmd, timeout=SUBPROC_TIMEOUT)
        requests.post("https://" + host_uri + "/upgrade/request/transfer/stop", json={"params": {
            "token": token,
        }}, timeout=TIMEOUT)
        return True

    def _get_contract(self):
        dbs_info_url = WEBSITE + '/odoo-enterprise/dbs-info'
        dbuuids = [self.db_uuid]
        if self.source_uuid:
            dbuuids.append(self.source_uuid)
        res = requests.post(dbs_info_url, data={'dbuuids': json.dumps(dbuuids)}, auth=SaasAuth(), timeout=TIMEOUT)
        dbs_info = {d['uuid']: d['code'] for d in res.json()} if res.status_code == 200 else {}
        my_code = dbs_info.get(self.db_uuid)
        source_code = dbs_info.get(self.source_uuid) if self.source_uuid else False
        return my_code or source_code or 'M8888-yyyy'


def get_upgrade_api(db, source_version, *args, **kwargs):
    token = kwargs.pop("token", None)
    return UpgradeAPI(db, *args, **kwargs)


def generate_token(request_id, days=3, timestep=0):
    """
    Compute a valid secret token for a request.
    This token is needed to download dumps which aren't supposed to be
    downloaded by end users, e.g. upgrade requests with `internal` mode/aim,
    requests which are still in testing etc.

    It act like a TOTP token within a `days` days window.
    You can get the previous/current/next token with the `timestep` argument
    """
    assert timestep in (-1, 0, 1)
    window = 24 * days

    time_frame = int(time.time() / 60 / 60 / window) + timestep
    vals = [str(time_frame), UPGRADE_KEYSALT, UPGRADE_LOCATION, str(request_id)]
    h = hashlib.md5(b':'.join(list(map(str.encode, vals))))
    token = base64.urlsafe_b64encode(h.digest()).decode('UTF-8')
    return token
